<?php

namespace Drupal\cl_components\Controller;

use Drupal\cl_components\Component\Component;
use Drupal\cl_components\Component\ComponentDiscovery;
use Drupal\cl_components\Component\ComponentMetadata;
use Drupal\cl_components\Exception\ComponentSyntaxException;
use Drupal\cl_components\Exception\InvalidComponentException;
use Drupal\cl_components\Parser\ComponentPhpFile;
use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\Tags;
use Drupal\Core\Controller\ControllerBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Controller for the registry.
 */
final class ComponentAudit extends ControllerBase {

  private ComponentDiscovery $discovery;

  /**
   * @param \Drupal\cl_components\Component\ComponentDiscovery $discovery
   */
  public function __construct(ComponentDiscovery $discovery) {
    $this->discovery = $discovery;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $discovery = $container->get(ComponentDiscovery::class);
    assert($discovery instanceof ComponentDiscovery);
    return new static($discovery);
  }

  /**
   * Render the registry.
   *
   * @return array
   *   The render array.
   */
  public function audit(): array {
    $components = $this->discovery->findAll();
    $paths = $this->discovery->findAllPaths();
    $build_component_card = fn(string $path) => $this->buildComponentCard($path, $components);
    return [
      'logs' => [
        '#prefix' => '<p>',
        '#markup' => $this->t('Remember to check your logs if you are having problems with your components.'),
        '#suffix' => '</p>',
      ],
      'components' => [
        '#type' => 'container',
        '#attributes' => ['class' => 'panel__container'],
        '#attached' => ['library' => ['cl_components/cl_registry']],
        'panels' => [
          '#theme' => 'item_list',
          '#title' => $this->t('Detected Components'),
          '#items' => array_map($build_component_card, $paths),
        ],
        'recommendations' => [
          '#prefix' => '<p>',
          '#markup' => $this->t('We recommend you to add a <code>README.md</code> file and a <code>thumbnail.png</code> to all your components. Other modules and JS tools may use them to provide additional insight.
'),
          '#suffix' => '</p>',
        ],
      ],
    ];
  }

  /**
   * Autocomplete endpoint for the search in the component selector.
   */
  public function autocomplete(Request $request, string $allowed, string $forbidden, string $types, string $statuses) {
    $filters = [
      'allowed' => Json::decode($allowed),
      'forbidden' => Json::decode($forbidden),
      'types' => Json::decode($types),
      'statuses' => Json::decode($statuses),
    ];
    // Get the typed string from the URL, if it exists.
    $input = $request->query->get('q');
    if (!$input) {
      return new JsonResponse([]);
    }
    $tag_list = Tags::explode($input);
    $typed_string = !empty($tag_list) ? mb_strtolower(array_pop($tag_list)) : '';
    $components = $this->discovery->findAllWithFilters(
      $filters['allowed'] ?? [],
      $filters['forbidden'] ?? [],
      $filters['types'] ?? [],
      $filters['statuses'] ?? [],
    );
    $filtered_components = array_filter(
      $components,
      static fn(Component $component) => str_contains($component->getId(), $typed_string)
        || str_contains($component->getMetadata()->getName(), $typed_string)
        || str_contains($component->getMetadata()->getGroup(), $typed_string)
    );
    $matches = array_map(
      static fn(Component $component) => $component->getId(),
      $filtered_components
    );
    // Refine matches with the typed string.
    return new JsonResponse(array_values($matches));
  }

  /**
   * Builds the render array for the component card.
   *
   * @param string $path
   *   The path to the component folder.
   * @param Component[] $components
   *   The components without duplications (for forked components).
   *
   * @return array[]
   *   The card render array.
   */
  private function buildComponentCard(string $path, array $components): array {
    try {
      $component = $this->discovery->instantiateComponent($path);
      $matches = array_filter($components, static fn (Component $cmp) => $cmp->getId() === $component->getId());
      // We should only have one match.
      $match = reset($matches);
      $is_forked = $match->getMetadata()->getPath() !== $component->getMetadata()->getPath();
    }
    catch (InvalidComponentException $e) {
      $component = NULL;
      $message = $this->t('Invalid component @path. Error: @error', [
        '@path' => $path,
        '@error' => $e->getMessage(),
      ]);
      return [
        'title' => [
          '#prefix' => '<h4>',
          '#markup' => $this->t('💥 Error in Component'),
          '#suffix' => '</h4>',
        ],
        'path' => [
          '#prefix' => '<pre>',
          '#markup' => $path,
          '#suffix' => '</pre>',
        ],
        'error' => [
          '#prefix' => '<p>',
          '#markup' => $e->getMessage(),
          '#suffix' => '</p>',
        ],
        '#wrapper_attributes' => ['class' => 'panel-item'],
      ];
    }
    $metadata = $component->getMetadata();
    $default_template = $component->getId() . '.twig';
    $template_message = in_array($default_template, $component->getTemplates())
      ? $this->t('✅ %default template is present', ['%default' => $default_template])
      : $this->t('❌ @id.twig is missing', ['@id' => $component->getId()]);
    $missing_variants = array_diff(
      [
        $default_template,
        ...array_map(static fn(string $variant) => sprintf('%s--%s.twig', $component->getId(), $variant), $component->getVariants()),
      ],
      $component->getTemplates()
    );
    $missing_variants = array_filter(
      $missing_variants,
      static fn(string $name) => $name !== $default_template
    );
    $variants_message = empty($missing_variants)
      ? $this->t('✅ All variants have their templates')
      : $this->t('❌ Missing templates: %variants', ['%variants' => implode(', ', $missing_variants)]);
    $assets = $this->discovery->discoverDistAssets($path);
    $assets_message = [
      '#theme' => 'item_list',
      '#items' => array_map(
        static fn(string $file) => [
          '#type' => 'html_tag',
          '#tag' => 'code',
          '#value' => preg_replace('@.*/' . $component->getId() . '/@', '', $file),
        ],
        [
          ...$assets['js'] ?? [],
          ...$assets['css'] ?? [],
        ]
      ),
    ];
    if (!$is_forked) {
      try {
        $hooks = $component->getMetadata()->getHooks();
      }
      catch (ComponentSyntaxException $e) {
        $hooks = [];
      }
      $hooks_message = [
        '#theme' => 'item_list',
        '#items' => array_map(
          static fn(string $hook) => [
            '#type' => 'html_tag',
            '#tag' => 'code',
            '#value' => $hook,
            '#attributes' => ['class' => in_array($hook, $hooks, TRUE) ? 'implemented' : 'not-implemented'],
          ],
          ComponentPhpFile::SUPPORTED_CL_HOOKS
        ),
      ];
    }
    else {
      $hooks_message = [
        '#type' => 'html_tag',
        '#tag' => 'em',
        '#value' => $this->t('Forked at <code>@filename</code>.', ['@filename' => $match->getMetadata()->getPath()]),
      ];
    }
    $card_build = [
      'title' => [
        '#theme' => 'cl_label_with_link',
        '#label' => $metadata->getName() . ($is_forked ? ' ❗' : ''),
        '#href' => '#' . $component->getId(),
      ],
      'description' => $metadata
        ->getDescription() === ComponentMetadata::DEFAULT_DESCRIPTION ? [] : [
        '#prefix' => '<p>',
        '#markup' => $metadata->getDescription(),
        '#suffix' => '</p>',
      ],
      'path' => [
        '#prefix' => '<pre>',
        '#markup' => $path,
        '#suffix' => '</pre>',
      ],
      'forked' => $is_forked ? [
        '#markup' => $this->t('When rendering this component by ID, this other component will render instead: <code>@filename</code>.', ['@filename' => $match->getMetadata()->getPath()]),
      ] : [],
      'table' => [
        '#theme' => 'table',
        '#header' => [
          $this->t('Metadata'),
          $this->t('Default Template'),
          $this->t('Variants'),
          $this->t('Assets'),
          $is_forked ? $this->t('🍴 Component Forked 🍴') : $this->t('Implemented Hooks'),
        ],
        '#rows' => [
          [
            $this->t('✅ All metadata is correct'),
            $template_message,
            $variants_message,
            ['data' => $assets_message],
            ['data' => $hooks_message],
          ],
        ],
      ],
      '#wrapper_attributes' => [
        'class' => ['panel-item', $is_forked ? 'forked' : 'not-forked'],
        'id' => $component->getId(),
      ],
    ];
    if (!$is_forked) {
      $this->moduleHandler()
        ->alter('cl_component_audit', $card_build, $component);
    }
    return $card_build;
  }

}
