<?php

namespace Drupal\cl_components\Element;

use Drupal\cl_components\Component\Component;
use Drupal\cl_components\Component\ComponentDiscovery;
use Drupal\cl_components\Component\ComponentMetadata;
use Drupal\cl_components\Exception\ComponentNotFoundException;
use Drupal\cl_components\Exception\InvalidComponentHookException;
use Drupal\Component\Serialization\Json;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\Element\FormElement;
use Drupal\Core\Render\Element\Radios;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a component selector element.
 *
 * @FormElement("cl_component_selector")
 */
class ComponentSelectorElement extends FormElement implements ContainerFactoryPluginInterface {

  protected const DEFAULT_STATUSES = [
    ComponentMetadata::COMPONENT_STATUS_READY,
    ComponentMetadata::COMPONENT_STATUS_BETA,
    ComponentMetadata::COMPONENT_STATUS_WIP,
  ];

  protected const DEFAULT_TYPES = [
    ComponentMetadata::COMPONENT_TYPE_ATOM,
    ComponentMetadata::COMPONENT_TYPE_MOLECULE,
    ComponentMetadata::COMPONENT_TYPE_ORGANISM,
  ];

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    protected ComponentDiscovery $discovery,
    protected FileUrlGeneratorInterface $fileUrlGenerator
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $discovery = $container->get(ComponentDiscovery::class);
    assert($discovery instanceof ComponentDiscovery);
    $file_url_generator = $container->get('file_url_generator');
    assert($file_url_generator instanceof FileUrlGeneratorInterface);
    return new static($configuration, $plugin_id, $plugin_definition, $discovery, $file_url_generator);
  }

  /**
   * @inheritDoc
   */
  public function getInfo() {
    return [
      '#title' => $this->t('CL Components'),
      '#description' => $this->t('Select the suitable CL component.'),
      '#filters' => [],
      '#process' => [[$this, 'populateOptions']],
      '#element_validate' => [[$this, 'validateExistingComponent']],
      '#theme_wrappers' => ['cl_component_selector', 'fieldset'],
      '#submit' => [[$this, 'submitForm']],
      '#input' => TRUE,
    ];
  }

  /**
   * Form API process callback.
   */
  public function populateOptions(
    array &$element,
    FormStateInterface $form_state,
    array &$complete_form
  ): array {
    $filters = $element['#filters'] ?? [];
    $components = $this->discovery->findAllWithFilters(
      $filters['allowed'] ?? [],
      $filters['forbidden'] ?? [],
      $filters['types'] ?? static::DEFAULT_TYPES,
      $filters['statuses'] ?? static::DEFAULT_STATUSES,
    );

    $options = array_map(
      static fn(Component $component) => $component->getMetadata()->getName(),
      $components
    );
    ksort($options);

    $element['#attributes']['class'] = ['cl-components--selector'];
    $default_value = $element['#default_value'] ?? NULL;
    $element += [
      '#attached' => ['library' => ['cl_components/options-filter']],
      'search' => [
        '#title' => $this->t('Search'),
        '#title_display' => 'hidden',
        '#type' => 'search',
        '#autocomplete_route_name' => 'cl_components.autocomplete',
        '#autocomplete_route_parameters' => [
          'allowed' => Json::encode($filters['allowed'] ?? []),
          'forbidden' => Json::encode($filters['forbidden'] ?? []),
          'types' => Json::encode($filters['types'] ?? static::DEFAULT_TYPES),
          'statuses' => Json::encode($filters['statuses'] ?? static::DEFAULT_STATUSES),
        ],
        '#default_value' => $default_value,
        '#placeholder' => $this->t('my-component'),
        '#size' => 30,
        '#description' => $this->t('Start typing the component name or group to reduce the list below.'),
        '#input' => FALSE,
      ],
      'radios' => [
        '#type' => 'radios',
        '#options' => $options,
        '#title' => $this->t('Components'),
        '#title_display' => 'hidden',
        '#default_value' => $default_value,
        '#process' => [
          [Radios::class, 'processRadios'],
          [$this, 'processRadios'],
        ],
        '#after_build' => [[$this, 'afterBuildRadios']],
        '#attributes' => ['class' => ['cl-components-selector--radios']],
        '#input' => FALSE,
        '#ajax' => $element['#ajax'] ?? FALSE,
      ],
    ];
    unset($element['#default_value'], $element['#ajax']);
    return $element;
  }

  public function afterBuildRadios(array $element, FormStateInterface $form_state): array {
    // This is a horrible hack, but I am at my wit's end.
    $value = $form_state->getValue($element['#parents']);
    if (isset($value)) {
      $form_state->setValue(array_slice($element['#parents'], 0, -1), $value);
    }
    return $element;
  }

  /**
   * @throws \Drupal\cl_components\Exception\ComponentNotFoundException
   */
  public function processRadios(array $element, FormStateInterface $form_state): array {
    unset($element['#value_callback']);
    $keys = Element::children($element);
    foreach ($keys as $key) {
      $element[$key]['#theme_wrappers'] = [
        'form_element__radio__cl_component',
        'form_element__radio',
      ];
      $id = $element[$key]['#return_value'];
      $component = $this->discovery->find($id);
      $element[$key]['#component'] = $component;
      $metadata = $component->getMetadata();
      $element[$key]['#description'] = $metadata->getDescription();
      $element[$key]['#value_callback'] = NULL;
      $element[$key]['#thumbnail_url'] = $metadata->getThumbnailPath() ? $this->fileUrlGenerator->generateAbsoluteString($metadata->getThumbnailPath()) : NULL;
      $element[$key]['#component_type'] = $metadata->getComponentType();
      $element[$key]['#component_status'] = $metadata->getStatus();
      $documentation = class_exists('\League\CommonMark\CommonMarkConverter')
        ? $metadata->getDocumentation()
        : $this->t('<em>Your README.md can be displayed here if you install: </em><code>composer require league/commonmark</code>');
      $element[$key]['#readme'] = $documentation;
      $element[$key]['#variants'] = $metadata->getVariants();
    }
    return $element;
  }

  /**
   * Validator for form element.
   */
  public function validateExistingComponent(array $element, FormStateInterface $form_state): void {
    $parents = $element['#parents'] ?? [];
    $value = $form_state->getValue($parents);
    if (!is_string($value)) {
      $value = NULL;
      $form_state->setValue($parents, NULL);
    }
    if (!$value) {
      return;
    }
    try {
      $this->discovery->find($value);
    }
    catch (ComponentNotFoundException $e) {
      $form_state->setError(
        $element['radios'],
        $this->t('Invalid component ID: @id', ['@id' => $value])
      );
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    return is_array($input) ? ($input['radios'] ?? NULL) : NULL;
  }

}
